from time import time

from typing_extensions import Self

from ._abstract import AbstractDomainUpdatesManager


class TimeoutDomainUpdatesManager(AbstractDomainUpdatesManager):
    def __init__(self, timeout: int):
        self._timeout = timeout

        self._timestamp = None

    async def __aenter__(self):
        self._timestamp = time()
        return self

    async def __aexit__(self, __exc_type, __exc_value, __traceback):
        self._timestamp = None

    @property
    async def has_changed(self) -> bool:
        if self._timestamp is None:
            raise AttributeError
        return time() - self._timestamp < self._timeout

    async def update(self) -> None:
        if self._timestamp is None:
            raise AttributeError
        self._timestamp = time()

    @classmethod
    def from_config(cls, config: dict) -> Self:
        return cls(config.get('timeout', 60))



__all__ = [
    'ComponentValueType',
    'IdentifyingPropertyType', 'PropertyType', 'RelationPropertyType',
    'RelationType',
    'AtomValueType'
]

from ._component import ComponentValueType
from ._property import IdentifyingPropertyType, PropertyType, RelationPropertyType
from ._relation import RelationType
from ._value import AtomValueType



from dataclasses import dataclass

from tp_interfaces.domain.abstract import AbstractRelationType
from ._relext import RelExtBasedType


@dataclass(frozen=True)
class RelationType(RelExtBasedType, AbstractRelationType):
    def __post_init__(self):
        RelExtBasedType.__post_init__(self)
        AbstractRelationType.__post_init__(self)

    async def _load_object(self) -> dict:
        return await self._api.relation_type_extras(self)



from dataclasses import dataclass

from tp_interfaces.domain.abstract import AbstractComponentValueType
from ._relext import RelExtBasedType


@dataclass(frozen=True)
class ComponentValueType(RelExtBasedType, AbstractComponentValueType):

    async def _load_object(self) -> dict:
        raise NotImplementedError



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field

from tp_interfaces.domain.abstract import AbstractRelExtBasedType, RelExtModel
from ._abstract import AbstractAdapterBasedType
from ._cached import CachedValue, cached


@dataclass(frozen=True)
class RelExtBasedType(AbstractAdapterBasedType, AbstractRelExtBasedType, metaclass=ABCMeta):
    _pretrained_relext_models: CachedValue[tuple[RelExtModel, ...]] = field(default_factory=CachedValue)

    def __post_init__(self):
        if not isinstance(self._pretrained_relext_models, CachedValue):
            raise ValueError
        AbstractAdapterBasedType.__post_init__(self)
        AbstractRelExtBasedType.__post_init__(self)

    @property
    @cached
    async def pretrained_relext_models(self) -> tuple[RelExtModel, ...]:
        type_ = await self._load_object()
        return tuple(RelExtModel(**model) for model in type_['pretrained_models'])

    @abstractmethod
    async def _load_object(self) -> dict:
        pass



from dataclasses import dataclass

from tp_interfaces.domain.abstract import AbstractLiteralValueType
from ._nerc import NERCBasedType


@dataclass(frozen=True)
class AtomValueType(NERCBasedType, AbstractLiteralValueType):
    def __post_init__(self):
        NERCBasedType.__post_init__(self)
        AbstractLiteralValueType.__post_init__(self)

    async def _load_object(
            self,
            regexp: bool = False,
            black_regexp: bool = False,
            pretrained_models: bool = False,
            dictionary: bool = False,
            black_list: bool = False
    ):
        return await self._api.literal_value_type_extras(self, regexp, black_regexp, pretrained_models, dictionary, black_list)



from dataclasses import dataclass

from tp_interfaces.domain.abstract import AbstractIdentifyingPropertyType, AbstractPropertyType, AbstractRelationPropertyType
from ._nerc import NERCBasedType
from ._relext import RelExtBasedType


@dataclass(frozen=True)
class PropertyType(RelExtBasedType, AbstractPropertyType):
    def __post_init__(self):
        RelExtBasedType.__post_init__(self)
        AbstractPropertyType.__post_init__(self)

    async def _load_object(self) -> dict:
        return await self._api.property_type_extras(self)


@dataclass(frozen=True)
class IdentifyingPropertyType(NERCBasedType, AbstractIdentifyingPropertyType):

    def __post_init__(self):
        NERCBasedType.__post_init__(self)
        AbstractIdentifyingPropertyType.__post_init__(self)

    async def _load_object(
            self,
            regexp: bool = False,
            black_regexp: bool = False,
            pretrained_models: bool = False,
            dictionary: bool = False,
            black_list: bool = False
    ):
        return await self._api.id_property_type_extras(self, regexp, black_regexp, pretrained_models, dictionary, black_list)


@dataclass(frozen=True)
class RelationPropertyType(RelExtBasedType, AbstractRelationPropertyType):

    def __post_init__(self):
        RelExtBasedType.__post_init__(self)
        AbstractPropertyType.__post_init__(self)

    async def _load_object(self) -> dict:
        return await self._api.relation_property_type_extras(self)



from typing import Type

from tp_interfaces.abstract import ModelTypeFactory
from ._abstract import AbstractDomainUpdatesManager


def _never() -> Type[AbstractDomainUpdatesManager]:
    from ._never import NeverDomainUpdatesManager
    return NeverDomainUpdatesManager


def _timeout() -> Type[AbstractDomainUpdatesManager]:
    from ._timeout import TimeoutDomainUpdatesManager
    return TimeoutDomainUpdatesManager


def _ask_kb() -> Type[AbstractDomainUpdatesManager]:
    from ._ask import AskKBDomainUpdatesManager
    return AskKBDomainUpdatesManager


MANAGERS = ModelTypeFactory({
    'never': _never,
    'timeout': _timeout,
    'ask kb': _ask_kb
})



from typing import Iterable

from tdm.abstract.datamodel import AbstractDomainType
from tdm.datamodel.domain import AccountType, Domain, PlatformType
from tdm.datamodel.values import StringValue

from tp_interfaces.domain.model.types import AtomValueType, IdentifyingPropertyType, PropertyType, RelationType
from ._common import STR_VALUE

_PLATFORM_TYPE_ENUM = (
    'government',
    'fileStorage',
    'database',
    'procurement',
    'review',
    'forum',
    'newsAggregator',
    'media',
    'blog',
    'messenger',
    'socialNetwork'
)

# VALUES

PLATFORM_VALUE = AtomValueType("platform_value", StringValue, id='platform_value', value_restriction=_PLATFORM_TYPE_ENUM)

# CONCEPTS
PLATFORM = PlatformType('platform', id='platform')
ACCOUNT = AccountType('account', id='account')

# RELATIONS
ACCOUNT_PLATFORM = RelationType("account_platform", ACCOUNT, PLATFORM, id="account_platform")

# PROPERTIES
ACCOUNT_KEY = IdentifyingPropertyType("Название", ACCOUNT, STR_VALUE, id="account_key")
ACCOUNT_NAME = PropertyType("account_name", ACCOUNT, STR_VALUE, id="account_name")
ACCOUNT_URL = PropertyType("account_url", ACCOUNT, STR_VALUE, id="account_url")

PLATFORM_KEY = IdentifyingPropertyType("Название", PLATFORM, STR_VALUE, id="platform_key")
PLATFORM_NAME = PropertyType("platform_name", PLATFORM, STR_VALUE, id="platform_name")
PLATFORM_URL = PropertyType("platform_url", PLATFORM, STR_VALUE, id="platform_url")
PLATFORM_TYPE = PropertyType("platform_type", PLATFORM, PLATFORM_VALUE, id="platform_type")

TYPES = (
    PLATFORM_VALUE, STR_VALUE,
    PLATFORM, ACCOUNT,
    ACCOUNT_PLATFORM,
    ACCOUNT_KEY, ACCOUNT_NAME, ACCOUNT_URL,
    PLATFORM_KEY, PLATFORM_NAME, PLATFORM_URL, PLATFORM_TYPE
)


def generate_types(domain: Domain) -> Iterable[AbstractDomainType]:
    return TYPES



import asyncio
from functools import wraps
from typing import Generic, TypeVar

T = TypeVar('T')


class CachedValue(Generic[T]):
    def __init__(self, value: T = None):
        self._value = value

    @property
    def empty(self) -> bool:
        return self._value is None

    @property
    def value(self) -> T:
        if self._value is None:
            raise AttributeError
        return self._value

    @value.setter
    def value(self, value: T) -> None:
        if self._value is not None:
            raise AttributeError
        self._value = value


def cached(f):
    name = f"_{f.__name__}"

    lock = asyncio.Lock()

    @wraps(f)
    async def eval_if_needed(self):
        attr = getattr(self, name)
        if not isinstance(attr, CachedValue):
            raise ValueError
        if not attr.empty:
            return attr.value
        async with lock:
            if not attr.empty:  # we should check it once again to guarantee single execution
                return attr.value
            attr.value = await f(self)
        return attr.value

    return eval_if_needed



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from itertools import chain

from tp_interfaces.domain.abstract import AbstractNERCBasedType, NERCRegexp
from ._abstract import AbstractAdapterBasedType
from ._cached import CachedValue, cached


@dataclass(frozen=True)
class NERCBasedType(AbstractAdapterBasedType, AbstractNERCBasedType, metaclass=ABCMeta):
    _regexp: CachedValue[tuple[NERCRegexp, ...]] = field(default_factory=CachedValue)
    _black_regexp: CachedValue[tuple[NERCRegexp, ...]] = field(default_factory=CachedValue)
    _pretrained_nerc_models: CachedValue[tuple[str, ...]] = field(default_factory=CachedValue)
    _dictionary: CachedValue[tuple[str, ...]] = field(default_factory=CachedValue)
    _black_list: CachedValue[tuple[str, ...]] = field(default_factory=CachedValue)

    def __post_init__(self):
        AbstractAdapterBasedType.__post_init__(self)
        AbstractNERCBasedType.__post_init__(self)
        if any(not isinstance(getattr(self, field), CachedValue) for field in [
            '_regexp', '_black_regexp', '_pretrained_nerc_models', '_dictionary', '_black_list'
        ]):
            raise ValueError

    @property
    @cached
    async def regexp(self) -> tuple[NERCRegexp, ...]:
        type_ = await self._load_object(regexp=True)
        return tuple(NERCRegexp(**data) for data in type_['regexp'])

    @property
    @cached
    async def black_regexp(self) -> tuple[NERCRegexp, ...]:
        type_ = await self._load_object(black_regexp=True)
        return tuple(NERCRegexp(**data) for data in type_['black_regexp'])

    @property
    @cached
    async def pretrained_nerc_models(self) -> tuple[str, ...]:
        type_ = await self._load_object(pretrained_models=True)
        return tuple(type_['pretrained_models'])

    @property
    @cached
    async def dictionary(self) -> tuple[str, ...]:
        type_ = await self._load_object(dictionary=True)
        return tuple(chain(type_.get('dictionary', ()), type_.get('names_dictionary', ())))

    @property
    @cached
    async def black_list(self) -> tuple[str, ...]:
        type_ = await self._load_object(black_list=True)
        return tuple(type_['black_list'])

    @abstractmethod
    async def _load_object(
            self,
            regexp: bool = False,
            black_regexp: bool = False,
            pretrained_models: bool = False,
            dictionary: bool = False,
            black_list: bool = False
    ):
        pass



from typing_extensions import Self

from ._abstract import AbstractDomainUpdatesManager


class NeverDomainUpdatesManager(AbstractDomainUpdatesManager):
    @property
    async def has_changed(self) -> bool:
        return False

    async def update(self) -> None:
        pass

    @classmethod
    def from_config(cls, config: dict) -> Self:
        return cls()

    async def __aexit__(self, __exc_type, __exc_value, __traceback):
        pass



from typing_extensions import Self

from talisman_api.api_client import APISchema, CompositeTalismanAPIClient
from talisman_api.api_client.composite import GQLClientConfig
from tie_domain.domain.api import TalismanDomainAPI
from ._abstract import AbstractDomainUpdatesManager


class AskKBDomainUpdatesManager(AbstractDomainUpdatesManager):
    def __init__(self, client: CompositeTalismanAPIClient):
        self._client = client
        self._api: TalismanDomainAPI | None = None

        self._timestamps: set[tuple] = set()

    async def __aenter__(self) -> Self:
        self._api = await TalismanDomainAPI.get_compatible_api(self._client)
        self._timestamps = await self._get_timestamps()
        return self

    async def __aexit__(self, __exc_type, __exc_value, __traceback):
        self._timestamps = set()
        self._api = None

    async def _get_timestamps(self) -> set[tuple]:
        return await self._api.domain_update_info()

    @property
    async def has_changed(self) -> bool:
        new_timestamps = await self._get_timestamps()
        return self._timestamps != new_timestamps

    async def update(self) -> None:
        self._timestamps = await self._get_timestamps()

    @classmethod
    def from_config(cls, config: dict) -> Self:
        return cls(CompositeTalismanAPIClient({APISchema(name): GQLClientConfig(**cfg) for name, cfg in config['client'].items()}))



from abc import ABCMeta, abstractmethod
from typing import AsyncContextManager

from typing_extensions import Self


class AbstractDomainUpdatesManager(AsyncContextManager, metaclass=ABCMeta):

    async def __aenter__(self):
        return self

    @property
    @abstractmethod
    async def has_changed(self) -> bool:
        pass

    @abstractmethod
    async def update(self) -> None:
        pass

    @classmethod
    @abstractmethod
    def from_config(cls, config: dict) -> Self:
        pass



__all__ = [
    'generate_types'
]

from itertools import chain
from typing import Iterable

from tdm.abstract.datamodel import AbstractDomainType
from tdm.datamodel.domain import Domain

from ._category import generate_types as generate_category_types
from ._platform import generate_types as generate_platform_types


def generate_types(domain: Domain) -> Iterable[AbstractDomainType]:
    return chain(
        generate_platform_types(domain),
        generate_category_types(domain)
    )



from abc import abstractmethod
from typing import AsyncIterator

from talisman_api import AbstractTalismanAPI
from tp_interfaces.domain.abstract import AbstractIdentifyingPropertyType, AbstractLiteralValueType, AbstractPropertyType, \
    AbstractRelationPropertyType, AbstractRelationType


class TalismanDomainAPI(AbstractTalismanAPI):

    @abstractmethod
    async def concept_types(self) -> AsyncIterator[dict]:
        pass

    @abstractmethod
    async def document_types(self) -> AsyncIterator[dict]:
        pass

    @abstractmethod
    async def literal_value_types(self) -> AsyncIterator[dict]:
        pass

    @abstractmethod
    async def composite_value_types(self) -> AsyncIterator[dict]:
        pass

    @abstractmethod
    async def relation_types(self) -> AsyncIterator[dict]:
        pass

    @abstractmethod
    async def property_types(self) -> AsyncIterator[dict]:
        pass

    @abstractmethod
    async def relation_property_types(self) -> AsyncIterator[dict]:
        pass

    @abstractmethod
    async def property_type_extras(self, prop_type: AbstractPropertyType) -> dict:
        pass

    @abstractmethod
    async def relation_property_type_extras(self, rel_prop_type: AbstractRelationPropertyType) -> dict:
        pass

    @abstractmethod
    async def relation_type_extras(self, rel_type: AbstractRelationType) -> dict:
        pass

    @abstractmethod
    async def id_property_type_extras(
            self,
            prop_type: AbstractIdentifyingPropertyType,
            regexp: bool = False,
            black_regexp: bool = False,
            pretrained_models: bool = False,
            dictionary: bool = False,
            black_list: bool = False
    ) -> dict:
        pass

    @abstractmethod
    async def literal_value_type_extras(
            self,
            value_type: AbstractLiteralValueType,
            regexp: bool = False,
            black_regexp: bool = False,
            pretrained_models: bool = False,
            dictionary: bool = False,
            black_list: bool = False
    ) -> dict:
        pass

    @abstractmethod
    async def domain_update_info(self) -> set[tuple[str, int]]:
        pass



from typing import AsyncIterator, Iterable

from tdm.abstract.datamodel.domain import AbstractLinkDomainType
from tdm.datamodel.domain.types import ConceptType, DocumentType

from talisman_api import APISchema
from tie_domain.domain.api import TalismanDomainAPI
from tp_interfaces.domain.abstract import AbstractIdentifyingPropertyType, AbstractLiteralValueType, AbstractPropertyType, \
    AbstractRelationPropertyType, AbstractRelationType


class TalismanDomainAPIImpl(TalismanDomainAPI):

    @classmethod
    def _required_apis(cls) -> Iterable[APISchema]:
        return (APISchema.KB_UTILS,)

    async def concept_types(self) -> AsyncIterator[dict]:
        async for i in self._client(APISchema.KB_UTILS).paginate_items('paginationConceptTypeIE'):
            yield i

    async def document_types(self) -> AsyncIterator[dict]:
        async for i in self._client(APISchema.KB_UTILS).paginate_items('paginationDocumentTypeIE'):
            yield i

    async def literal_value_types(self) -> AsyncIterator[dict]:
        async for i in self._client(APISchema.KB_UTILS).paginate_items('paginationValueTypeIE'):
            yield i

    async def composite_value_types(self) -> AsyncIterator[dict]:
        async for i in self._client(APISchema.KB_UTILS).paginate_items('paginationCompositeValueTypeIE'):
            yield i

    async def relation_types(self) -> AsyncIterator[dict]:

        async for i in self._client(APISchema.KB_UTILS).paginate_items('paginationRelationTypeIE'):
            yield i

    async def property_types(self) -> AsyncIterator[dict]:
        async for i in self._client(APISchema.KB_UTILS).get_all_items('paginationConceptPropertyTypeIE'):  # see TKL-2966
            yield i

    async def relation_property_types(self) -> AsyncIterator[dict]:
        async for i in self._client(APISchema.KB_UTILS).get_all_items('paginationRelationPropertyTypeIE'):  # see TKL-2966
            yield i

    async def property_type_extras(self, prop_type: AbstractPropertyType) -> dict:
        return await self._relext_based_type(prop_type, 'propertyTypeExtrasIE')

    async def relation_property_type_extras(self, rel_prop_type: AbstractRelationPropertyType) -> dict:
        return await self._relext_based_type(rel_prop_type, 'relationPropertyTypeExtrasIE')

    async def relation_type_extras(self, rel_type: AbstractRelationType) -> dict:
        return await self._relext_based_type(rel_type, 'relationTypeExtrasIE')

    async def _relext_based_type(self, type_: AbstractLinkDomainType, operation: str) -> dict:
        return await self._load_object(operation, {'source_type_id': type_.source.id, 'name': type_.id})

    async def id_property_type_extras(
            self,
            prop_type: AbstractIdentifyingPropertyType,
            regexp: bool = False,
            black_regexp: bool = False,
            pretrained_models: bool = False,
            dictionary: bool = False,
            black_list: bool = False
    ) -> dict:
        if isinstance(prop_type.source, ConceptType):
            operation = 'conceptTypeExtrasIE'
        elif isinstance(prop_type.source, DocumentType):
            operation = 'documentTypeExtrasIE'
        else:
            raise ValueError
        variables = {
            'name': prop_type.source.id,
            'regexp': regexp,
            'black_regexp': black_regexp,
            'pretrained_models': pretrained_models,
            'dictionary': dictionary,
            'black_list': black_list
        }
        return await self._load_object(operation, variables)

    async def literal_value_type_extras(
            self,
            value_type: AbstractLiteralValueType,
            regexp: bool = False,
            black_regexp: bool = False,
            pretrained_models: bool = False,
            dictionary: bool = False,
            black_list: bool = False
    ) -> dict:
        variables = {
            'name': value_type.id,
            'regexp': regexp,
            'black_regexp': black_regexp,
            'pretrained_models': pretrained_models,
            'dictionary': dictionary,
            'black_list': black_list
        }
        return await self._load_object('valueTypeExtrasIE', variables)

    async def domain_update_info(self) -> set[tuple[str, int]]:
        return {
            (update['name'], update['updateDate'])
            for update in (await self._client(APISchema.KB_UTILS).execute('domainUpdateInfoIE'))['info']
        }

    async def _load_object(self, operation_name: str, variables: dict) -> dict:
        result = await self._find_object(operation_name, variables)
        if result is None:
            raise ValueError(f"no type found [operation: {operation_name}, variables: {variables}]")
        return result

    async def _find_object(self, operation_name: str, variables: dict) -> dict | None:
        if 'offset' in variables:
            raise ValueError
        offset = 0
        total = 1
        while offset < total:
            vs = {"offset": offset}
            vs.update(variables)
            result = (await self._client(APISchema.KB_UTILS).execute(operation_name, variables=vs))['result']
            if result['total'] == 0:
                return None
            total = result['total']
            for item in result['list']:
                if item['id'] == vs['name']:
                    return item
            offset += len(result['list'])



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("VERSION", "r", encoding="utf-8") as f:
    version = f.read()

setup(
    name='talisman-domain',
    version=version,
    description='Talisman domain implementations',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='ISPRAS Talisman NLP team',
    author_email='modis@ispras.ru',
    maintainer='Vladimir Mayorov',
    maintainer_email='vmayorov@ispras.ru',
    packages=find_packages(include=['tie_domain', 'tie_domain.*']),
    install_requires=[
        'talisman-interfaces>=0.8,<0.12', 'talisman-api~=1.3', 'talisman-dm>=1.3.1,<2',
        'typing_extensions>=4.0.0'
    ],
    entry_points={
        'talisman.plugins': ['domain = tie_domain']
    },
    data_files=[('', ['VERSION'])],
    package_data={
        'tie_domain.domain.api._impl': ['graphql/**/*']
    },
    python_requires='>=3.10',
    license='Apache Software License',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.11',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License'
    ]
)



from abc import ABCMeta
from dataclasses import dataclass

from tdm.abstract.datamodel import AbstractDomainType

from tie_domain.domain.api import TalismanDomainAPI


@dataclass(frozen=True)
class AbstractAdapterBasedType(AbstractDomainType, metaclass=ABCMeta):
    _api: TalismanDomainAPI = None

    def __post_init__(self):
        if self._api is None:
            raise ValueError



import logging
import os
from typing import Iterable, TypeVar

from tdm.abstract.datamodel import AbstractDomainType
from tdm.datamodel.domain import CompositeValueType, ConceptType, DocumentType, Domain

from talisman_api.api_client import APISchema, CompositeTalismanAPIClient
from talisman_api.api_client.client import GQLClientConfig
from tp_interfaces.domain.hooks import DOMAIN_CHANGE_HOOKS
from tp_interfaces.domain.interfaces import AbstractDomainChangeHook, DomainProducer
from tp_interfaces.domain.value import ValueTypes
from tp_interfaces.logging.time import log_time
from ._types import AtomValueType, ComponentValueType, IdentifyingPropertyType, PropertyType, RelationPropertyType, RelationType
from .api import TalismanDomainAPI
from .updates_manager import AbstractDomainUpdatesManager, MANAGERS

logger = logging.getLogger(__name__)

_SYNTHETIC_TYPES = bool(os.getenv('SYNTHETIC_TYPES', True))

_T = TypeVar('_T', bound=AbstractDomainType)


class TalismanDomainProducer(DomainProducer):
    def __init__(
            self, client: CompositeTalismanAPIClient, updates_manager: AbstractDomainUpdatesManager,
            hooks: Iterable[AbstractDomainChangeHook] = tuple()
    ):
        super().__init__(hooks)
        self._client = client
        self._api: TalismanDomainAPI | None = None
        self._updates_manager = updates_manager

    async def __aenter__(self):
        self._api = await TalismanDomainAPI.get_compatible_api(self._client)
        await self._updates_manager.__aenter__()
        return self

    async def __aexit__(self, *exc):
        await self._updates_manager.__aexit__(*exc)
        self._api = None

    @log_time(logger=logger)
    async def has_changed(self) -> bool:
        return await self._updates_manager.has_changed

    @log_time(logger=logger)
    async def _get_domain(self) -> Domain:
        await self._updates_manager.update()  # first we notify manager, that we reload domain

        types: dict[str, AbstractDomainType] = {}

        async for concept in self._api.concept_types():
            types[concept['id']] = ConceptType(name=concept['name'], id=concept['id'])

        async for document in self._api.document_types():
            types[document['id']] = DocumentType(name=document['name'], id=document['id'])

        async for value in self._api.literal_value_types():
            types[value['id']] = AtomValueType(
                name=value['name'],
                value_type=ValueTypes.get(value['valueType'], value['valueType']),
                id=value['id'],
                value_restriction=tuple(value['valueRestriction']),
                _api=self._api
            )

        async for composite in self._api.composite_value_types():
            c = CompositeValueType(name=composite['name'], id=composite['id'])
            types[c.id] = c

            for component in composite['componentValueTypes']:
                types[component['id']] = ComponentValueType(
                    name=component['name'],
                    source=c,
                    target=types[component['valueType']['id']],
                    id=component['id'],
                    isRequired=component['isRequired'],
                    _api=self._api
                )
        async for relation in self._api.relation_types():
            types[relation['id']] = RelationType(
                name=relation['name'],
                source=types[relation['source']['id']],
                target=types[relation['target']['id']],
                id=relation['id'],
                directed=relation['isDirected'],
                _api=self._api
            )
        async for prop in self._api.property_types():
            if prop['isIdentifying']:
                types[prop['id']] = IdentifyingPropertyType(
                    name=prop['name'],
                    source=types[prop['source']['id']],
                    target=types[prop['target']['id']],
                    id=prop['id'],
                    _api=self._api
                )
            else:
                types[prop['id']] = PropertyType(
                    name=prop['name'],
                    source=types[prop['source']['id']],
                    target=types[prop['target']['id']],
                    id=prop['id'],
                    _api=self._api
                )
        async for prop in self._api.relation_property_types():
            types[prop['id']] = RelationPropertyType(
                name=prop['name'],
                source=types[prop['source']['id']],
                target=types[prop['target']['id']],
                id=prop['id'],
                _api=self._api
            )

        final_types = set(types.values())

        if _SYNTHETIC_TYPES:
            from ._synthetic import generate_types
            final_types.update(generate_types(Domain(final_types)))
        return Domain(final_types)

    @classmethod
    def from_config(cls, config: dict) -> 'TalismanDomainProducer':
        hooks = []
        for hook_cfg in config.get('hooks', []):
            if isinstance(hook_cfg, str):
                name, cfg = hook_cfg, {}
            elif isinstance(hook_cfg, dict):
                name, cfg = hook_cfg['name'], hook_cfg.get('config', {})
            else:
                raise ValueError
            hooks.append(DOMAIN_CHANGE_HOOKS[name].from_config(cfg))
        updates_cfg = config.get('updates', {})
        updates_manager = MANAGERS[updates_cfg.get('strategy', 'never')].from_config(updates_cfg.get('config', {}))
        client = CompositeTalismanAPIClient({APISchema(name): GQLClientConfig(**cfg) for name, cfg in config['adapter'].items()})
        return cls(client, updates_manager, hooks)



from typing import Type

from tp_interfaces.domain.interfaces import DomainProducer


def _get_domain() -> Type[DomainProducer]:
    from tie_domain.domain.producer import TalismanDomainProducer
    return TalismanDomainProducer


DOMAIN_FACTORY = _get_domain



from typing import Iterable

from tdm.abstract.datamodel import AbstractDomainType
from tdm.datamodel.domain import DocumentType, Domain

from tp_interfaces.domain.model.types import PropertyType
from ._common import STR_VALUE


def generate_types(domain: Domain) -> Iterable[AbstractDomainType]:
    result = []
    for doc_type in domain.get_types(DocumentType):
        result.append(PropertyType('Рубрика', doc_type, STR_VALUE, id=f'{doc_type.id}:category'))
    if result:
        result.append(STR_VALUE)
    return result
